# coding: utf-8
"""
Parse BSC's SLURM Simulator and

 - Plot utilization and required resouces per instant
 - Share of backfilled jobs

BSC-CNS - Earth Sciences
Manuel Giménez de Castro Marciani
2023
"""
import pandas as pd
import sys 
import matplotlib.pyplot as plt
import numpy as np
import pyswf.pyswf as pyswf

def workflow_statistics(trace_dataframe: pd.DataFrame, target_uid:int, machine:str, workflow_name:str, wrapped:str, trace_path:str) -> None:
    """
    Generates run statistics of the target workflow
    :trace_dataframe: dataframe representation of the simulator output trace
    :target_uid: uid that launched the simulation
    :machine: name of the machine that was simulated
    :workflow_name: name of the workflow
    :wrapped: if the workflow was wrapped
    :trace_path: path to save the figure
    """
    # TARGET WORKFLOW
    # manuel: slowfown for a horizontal workflow is not really well defined.
    with open(trace_path+f"{machine}_{workflow_name}_{wrapped}.csv", "w") as wf_outputfile:
        target_workflow = trace_dataframe[trace_dataframe["user_id"] == target_uid]
        if "horizontal" in workflow_name:
            wf_runtime = target_workflow["runtime"].iloc[0]  # we assume every job has the same runtime
        else:  # vertical
            wf_runtime = target_workflow["runtime"].sum()
        wf_total_wait_time = target_workflow["wait_time"].sum()
        wf_total_time = target_workflow["end_time"].max() - target_workflow["submit_time"].min()
        wf_slowdown = wf_total_time/wf_runtime
        wf_outputfile.write(f"simulation, total_runtime, total_time, slowdown, total_wait_time, runtime_diff\n")
        wf_outputfile.write(f"{machine}_{workflow_name}_{wrapped}, {wf_runtime}, {wf_total_time}, {wf_slowdown}, {wf_total_wait_time}, {wf_total_time - wf_runtime}\n")


def utilization_plot(trace_dataframe: pd.DataFrame, target_uid:int, machine:str, workflow_name:str, wrapped:str,
                     trace_path:str, total_cpus: int = 0, xbot: int = 0, xtop: int = 0) -> None:
    """
    Plots the utilization and requested plot of the data.
    :trace_dataframe: dataframe representation of the simulator output trace
    :target_uid: uid of the user that launched the simulation
    :machine: name of the machine that was simulated
    :workflow_name: name of the workflow
    :wrapped: if the workflow was wrapped
    :total_cpus: total available cpus in the machine. If set to 0, absolute values will be plotted
    :xbot: left limit of the window that will be saved
    :xtop: right limit of the window that will be saved
    :trace_path: path to save the figure
    """

    # only compute the utilization where there were changes, and the previous instant.
    time_steps = list(trace_dataframe["start_time"]) + list(trace_dataframe["end_time"]) + list(trace_dataframe["submit_time"]) + list(trace_dataframe["start_time"]-1) + list(trace_dataframe["end_time"]-1) + list(trace_dataframe["submit_time"]-1) 
    time_steps = list(set(time_steps))  # remove duplicates
    time_steps.sort()
    
    utilization = []
    requested = []

    for time in time_steps:
        active_jobs = trace_dataframe[(trace_dataframe["start_time"] <= time) & (trace_dataframe["end_time"] > time)]
        utilization.append(active_jobs["proc_count"].sum())
        queueing_jobs = trace_dataframe[(trace_dataframe["submit_time"] <= time) & (trace_dataframe["start_time"] > time)]
        requested.append(queueing_jobs["proc_count"].sum())
   
    if total_cpus != 0:
        plt.plot(time_steps, np.array(utilization)/total_cpus, label="util")
        plt.plot(time_steps, np.array(requested)/total_cpus, label="req")
    else:
        plt.plot(time_steps, utilization, label="util")
        plt.plot(time_steps, requested, label="req")

    # draw vertical lines where the workflow was launched
    for i, wf_submit_time in enumerate(trace_dataframe[trace_dataframe["user_id"] == target_uid]["submit_time"]):
        if i == 0:
            plt.axvline(wf_submit_time, color="r", label=f"WF job")
        else:
            plt.axvline(wf_submit_time, color="r")

    plt.title(f"Simulated Utilized and Requested resources\ntrace with {workflow_name} {wrapped} at {machine}")
    plt.legend()

    if xbot != 0 or xtop != 0:
        plt.xlim(xbot, xtop)

    plt.savefig(trace_path+f"{machine}_{workflow_name}_{wrapped}_util_req.png")
    plt.clf()  # make sure to clear the canvas always

def backfill_plot(trace_dataframe: pd.DataFrame, target_uid:int, machine:str, workflow_name:str, wrapped:str, trace_path:str):
    """
    Plots the utilization and requested plot of the data.
    :trace_dataframe: dataframe representation of the simulator output trace
    :target_uid: uid of the user that launched the simulation
    :machine: name of the machine that was simulated
    :workflow_name: name of the workflow
    :wrapped: if the workflow was wrapped
    :trace_path: path to save the figure
    """

    # Backfilled PLOT
    total = trace_dataframe["backfill"].describe()[0]
    count = trace_dataframe["backfill"].describe()[3]
    common = trace_dataframe["backfill"].describe()[2]
    # super sketchy way of picking the other label
    uncommon = (set(trace_dataframe["backfill"].unique())-{common}).pop()
    plt.pie([count, total-count], labels=[common, uncommon])
    plt.title(f"Proportion of jobs backfilled \ntrace with {workflow_name} {wrapped} at {machine}")
    plt.savefig(trace_path+f"{machine}_{workflow_name}_{wrapped}_backfill.png")
    plt.clf()


if __name__ == "__main__":

    if len(sys.argv) < 5:
        sys.exit("usage: python3 parser_plotter.py <input_file> <target_uid> <machine> <workflow_name> <wrapped> {total_cpus} {xbot} {xtop}\n\ninput_file: slurm trace output file\ntarget_uid: uid of the user that launched the job\nmachine: name of the machine that was simulated\nworkflow_name: target workflow\nwrapped_flag: if the workflow was wrapped or not\ntotal_cpus: optional parameter with the total number of available cpus in the machine\nxbot: lower corner of the window to be plotted in the utilization plot\nxtop: top corner of the window to be plotted in the utilization plot\n")
    
    trace_file = sys.argv[1]
    trace_path = trace_file.rsplit("/",1)[0]+"/"
    target_uid = int(sys.argv[2])
    machine = sys.argv[3]
    workflow_name = sys.argv[4]
    wrapped = sys.argv[5]

    # is try-catch a good idea for optional parameters?
    try:
        total_cpus = int(sys.argv[6])
    except IndexError:
        total_cpus = 0

    try:
        xbot = int(sys.argv[7])
    except IndexError:
        xbot = 0

    try:
        xtop = int(sys.argv[8])
    except IndexError:
        xtop = 0

    trace_out = pyswf.SWFtrace.from_bsctrace(trace_file)

    # output wait time stats to csv
    trace_out.job_list["wait_time"].describe().to_csv(f'{machine}_{workflow_name}_{wrapped}.csv')

    workflow_statistics(trace_out.job_list, target_uid, machine, workflow_name, wrapped, trace_path)
    utilization_plot(trace_out.job_list, target_uid, machine, workflow_name, wrapped, trace_path, total_cpus, xbot, xtop)
    backfill_plot(trace_out.job_list, target_uid, machine, workflow_name, wrapped, trace_path)
